__author__ = 'mnowotka'

import pytz
from datetime import datetime, timedelta
from chembl_beaker.beaker.throttle.backends.base import BaseThrottle
from chembl_beaker.beaker.cache import cache

#-----------------------------------------------------------------------------------------------------------------------

class CacheThrottle(BaseThrottle):

    def __init__(self):
        BaseThrottle.__init__(self)
        if not cache:
            err = "CacheThrottle class can't work without cache..."
            print err
            raise Exception(err)

    def get_remaining_rates(self, identifier, type='IP'):
        start_usage = (None, self.hourly_rate_limit, self.daily_rate_limit) if type == 'IP' else \
                                                        (None, self.key_hourly_rate_limit, self.key_daily_rate_limit)
        usage = cache.get(identifier, start_usage)
        return usage[1], usage[2]

    def accessed(self, identifier, type='IP'):
        """
        Handles recording the user's access.

        Stores the current timestamp in the "accesses" list within the cache.
        """
        start_usage = (None, self.hourly_rate_limit, self.daily_rate_limit) if type == 'IP' else \
                                                        (None, self.key_hourly_rate_limit, self.key_daily_rate_limit)
        _, hourly_rate_limit, daily_rate_limit = start_usage
        usage = cache.get(identifier, start_usage)
        now = datetime.utcnow().replace(tzinfo=pytz.utc)
        last_usage, old_hourly_rate_limit, old_daily_rate_limit = usage
        if not last_usage:
            cache.set(identifier, (now, max(old_hourly_rate_limit-1, 0), max(old_daily_rate_limit-1, 0)))
        else:
            if last_usage + timedelta(hours=1) >= now and now.hour == last_usage.hour:
                new_hourly_rate_limit = max(old_hourly_rate_limit-1, 0)
            else:
                new_hourly_rate_limit = hourly_rate_limit
            if last_usage + timedelta(days=1) >= now and now.day == last_usage.day:
                new_daily_rate_limit = max(old_daily_rate_limit-1, 0)
            else:
                new_daily_rate_limit = daily_rate_limit
            cache.set(identifier, (now, new_hourly_rate_limit, new_daily_rate_limit))

#-----------------------------------------------------------------------------------------------------------------------



# $Id: cairoCanvas.py 11930 2014-01-24 07:00:08Z landrgr1 $
#
#  Copyright (C) 2008 Greg Landrum
#  Copyright (C) 2009 Uwe Hoffmann
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#

try:
  import cairo
except ImportError:
  import cairocffi as cairo
if not hasattr(cairo.ImageSurface,'get_data') and \
   not hasattr(cairo.ImageSurface,'get_data_as_rgba'):
  raise ImportError('cairo version too old')


import math
import os,re
import array
if not os.environ.has_key('RDK_NOPANGO'):
  try:
    import pangocairo
  except ImportError:
    pangocairo=None
  try:
    import pango
  except ImportError:
    pango=None
else:
  pango=None
  pangocairo=None

from canvasbase import CanvasBase
try:
  import Image
except ImportError:
  from PIL import Image

scriptPattern=re.compile(r'\<.+?\>')
  
class Canvas(CanvasBase):
  def __init__(self,
               image=None,  # PIL image
               size=None,
               ctx=None,
               imageType=None, # determines file type
               fileName=None,  # if set determines output file name
               ):
    """
    Canvas can be used in four modes:
    1) using the supplied PIL image
    2) using the supplied cairo context ctx
    3) writing to a file fileName with image type imageType
    4) creating a cairo surface and context within the constructor
    """
    self.image=None
    self.imageType=imageType
    if image is not None:
      try:
      	imgd = image.tostring("raw","BGRA")
      except SystemError:
        r,g,b,a = image.split()
        imgd = Image.merge("RGBA",(b,g,r,a)).tostring("raw","RGBA")
   
      a = array.array('B',imgd)
      stride=image.size[0]*4
      surface = cairo.ImageSurface.create_for_data (
        a, cairo.FORMAT_ARGB32,
        image.size[0], image.size[1], stride)
      ctx = cairo.Context(surface)
      size=image.size[0], image.size[1]
      self.image=image
    elif ctx is None and size is not None:
      if cairo.HAS_PDF_SURFACE and imageType == "pdf":
        surface = cairo.PDFSurface (fileName, size[0], size[1])
      elif cairo.HAS_SVG_SURFACE and imageType == "svg":
        surface = cairo.SVGSurface (fileName, size[0], size[1])
      elif cairo.HAS_PS_SURFACE and imageType == "ps":
        surface = cairo.PSSurface (fileName, size[0], size[1])
      elif imageType == "png":
        surface = cairo.ImageSurface (cairo.FORMAT_ARGB32, size[0], size[1])
      else:
        raise ValueError("Unrecognized file type. Valid choices are pdf, svg, ps, and png")
      ctx = cairo.Context(surface)
      ctx.set_source_rgb(1,1,1)
      ctx.paint()
    else:
      surface=ctx.get_target()
      if size is None:
        try:
          size=surface.get_width(),surface.get_height() 
        except AttributeError:
          size=None
    self.ctx=ctx
    self.size=size
    self.surface=surface
    self.fileName=fileName

  def flush(self):
    """temporary interface, must be splitted to different methods,
    """
    if self.fileName and self.imageType=='png':
      self.surface.write_to_png(self.fileName)
    elif self.image is not None:
      # on linux at least it seems like the PIL images are BGRA, not RGBA:
      if hasattr(self.surface,'get_data'):
        self.image.fromstring(self.surface.get_data(),
                              "raw","BGRA",0,1)
      else:
        self.image.fromstring(self.surface.get_data_as_rgba(),
                              "raw","RGBA",0,1)
      self.surface.finish()
    elif self.imageType == "png":
      if hasattr(self.surface,'get_data'):
        buffer=self.surface.get_data()
      else:
        buffer=self.surface.get_data_as_rgba()
      return buffer

  def _doLine(self, p1, p2, **kwargs):
    if kwargs.get('dash',(0,0)) == (0,0):
      self.ctx.move_to(p1[0],p1[1])
      self.ctx.line_to(p2[0],p2[1])
    else:
      dash = kwargs['dash']
      pts = self._getLinePoints(p1,p2,dash)

      currDash = 0
      dashOn = True
      while currDash<(len(pts)-1):
        if dashOn:
          p1 = pts[currDash]
          p2 = pts[currDash+1]
          self.ctx.move_to(p1[0],p1[1])
          self.ctx.line_to(p2[0],p2[1])
        currDash+=1
        dashOn = not dashOn

  def addCanvasLine(self,p1,p2,color=(0,0,0),color2=None,**kwargs):
    self.ctx.set_line_width(kwargs.get('linewidth',1))
    if color2 and color2!=color:
      mp = (p1[0]+p2[0])/2.,(p1[1]+p2[1])/2.
      self.ctx.set_source_rgb(*color)
      self._doLine(p1,mp,**kwargs)
      self.ctx.stroke()
      self.ctx.set_source_rgb(*color2)
      self._doLine(mp,p2,**kwargs)
      self.ctx.stroke()
    else:
      self.ctx.set_source_rgb(*color)
      self._doLine(p1,p2,**kwargs)
      self.ctx.stroke()

  def _addCanvasText1(self,text,pos,font,color=(0,0,0),**kwargs):
    if font.weight=='bold':
      weight=cairo.FONT_WEIGHT_BOLD
    else:
      weight=cairo.FONT_WEIGHT_NORMAL
    self.ctx.select_font_face(font.face,
                                cairo.FONT_SLANT_NORMAL,
                                weight)
    text = scriptPattern.sub('',text)
    self.ctx.set_font_size(font.size)
    w,h=self.ctx.text_extents(text)[2:4]
    bw,bh=w+h*0.4,h*1.4
    offset = w*pos[2]
    dPos = pos[0]-w/2.+offset,pos[1]+h/2.
    self.ctx.set_source_rgb(*color)
    self.ctx.move_to(*dPos)
    self.ctx.show_text(text)

    if 0:
      self.ctx.move_to(dPos[0],dPos[1])
      self.ctx.line_to(dPos[0]+bw,dPos[1])
      self.ctx.line_to(dPos[0]+bw,dPos[1]-bh)
      self.ctx.line_to(dPos[0],dPos[1]-bh)
      self.ctx.line_to(dPos[0],dPos[1])
      self.ctx.close_path()
      self.ctx.stroke()

    return (bw,bh,offset)
  def _addCanvasText2(self,text,pos,font,color=(0,0,0),**kwargs):
    if font.weight=='bold':
      weight=cairo.FONT_WEIGHT_BOLD
    else:
      weight=cairo.FONT_WEIGHT_NORMAL
    self.ctx.select_font_face(font.face,
                                cairo.FONT_SLANT_NORMAL,
                                weight)
    orientation=kwargs.get('orientation','E')
    cctx=pangocairo.CairoContext(self.ctx)

    plainText = scriptPattern.sub('',text)
    measureLout = cctx.create_layout()
    measureLout.set_alignment(pango.ALIGN_LEFT)
    measureLout.set_markup(plainText)

    lout = cctx.create_layout()
    lout.set_alignment(pango.ALIGN_LEFT)
    lout.set_markup(text)

    # for whatever reason, the font size using pango is larger
    # than that w/ default cairo (at least for me)
    fnt = pango.FontDescription('%s %d'%(font.face,font.size*.8))
    lout.set_font_description(fnt)
    measureLout.set_font_description(fnt)

    # this is a bit kludgy, but empirically we end up with too much
    # vertical padding if we use the text box with super and subscripts
    # for the measurement. 
    iext,lext=measureLout.get_pixel_extents()
    iext2,lext2=lout.get_pixel_extents()
    w=lext2[2]-lext2[0]
    h=lext[3]-lext[1]
    pad = [h*.2,h*.3]
    # another empirical correction: labels draw at the bottom
    # of bonds have too much vertical padding
    if orientation=='S':
      pad[1] *= 0.5
    bw,bh=w+pad[0],h+pad[1]
    offset = w*pos[2]
    if 0:
      if orientation=='W':
        dPos = pos[0]-w+offset,pos[1]-h/2.
      elif orientation=='E':
        dPos = pos[0]-w/2+offset,pos[1]-h/2.
      else:
        dPos = pos[0]-w/2+offset,pos[1]-h/2.
      self.ctx.move_to(dPos[0],dPos[1])
    else:
      dPos = pos[0]-w/2.+offset,pos[1]-h/2.
      self.ctx.move_to(dPos[0],dPos[1])
      
    self.ctx.set_source_rgb(*color)
    cctx.update_layout(lout)
    cctx.show_layout(lout)

    if 0:
      self.ctx.move_to(dPos[0],dPos[1])
      self.ctx.line_to(dPos[0]+bw,dPos[1])
      self.ctx.line_to(dPos[0]+bw,dPos[1]+bh)
      self.ctx.line_to(dPos[0],dPos[1]+bh)
      self.ctx.line_to(dPos[0],dPos[1])
      self.ctx.close_path()
      self.ctx.stroke()
    

    return (bw,bh,offset)

  def addCanvasText(self,text,pos,font,color=(0,0,0),**kwargs):
    if pango is not None and pangocairo is not None:
      textSize = self._addCanvasText2(text,pos,font,color,**kwargs)
    else:
      textSize = self._addCanvasText1(text,pos,font,color,**kwargs)
    return textSize
    
  def addCanvasPolygon(self,ps,color=(0,0,0),fill=True,stroke=False,**kwargs):
    if not fill and not stroke: return
    dps = []
    self.ctx.set_source_rgb(*color)
    self.ctx.move_to(ps[0][0],ps[0][1])
    for p in ps[1:]:
      self.ctx.line_to(p[0],p[1])
    self.ctx.close_path()
    if stroke:
      if fill:
        self.ctx.stroke_preserve()
      else:
        self.ctx.stroke()
    if fill:
      self.ctx.fill()

  def addCanvasDashedWedge(self,p1,p2,p3,dash=(2,2),color=(0,0,0),
                           color2=None,**kwargs):
    self.ctx.set_line_width(kwargs.get('linewidth',1))
    self.ctx.set_source_rgb(*color)
    dash = (3,3)
    pts1 = self._getLinePoints(p1,p2,dash)
    pts2 = self._getLinePoints(p1,p3,dash)

    if len(pts2)<len(pts1): pts2,pts1=pts1,pts2

    for i in range(len(pts1)):
      self.ctx.move_to(pts1[i][0],pts1[i][1])
      self.ctx.line_to(pts2[i][0],pts2[i][1])
    self.ctx.stroke()

  def addCircle(self,center,radius,color=(0,0,0),fill=True,stroke=False,alpha=1.0,
                **kwargs):
    if not fill and not stroke: return
    dps = []
    #import pdb; pdb.set_trace();
    self.ctx.set_source_rgba(color[0],color[1],color[2],alpha)
    self.ctx.arc(center[0],center[1],radius,0,2.*math.pi)
    self.ctx.close_path()
    if stroke:
      if fill:
        self.ctx.stroke_preserve()
      else:
        self.ctx.stroke()
    if fill:
      self.ctx.fill()



from chembl_beaker.beaker import config
from chembl_beaker.beaker.utils import import_class
from chembl_beaker.beaker.throttle.backends.base import BaseThrottle

throttle = None

throttle_class = config.get('throttle_backend', 'chembl_beaker.beaker.throttle.backends.cacheThrottle.CacheThrottle')

if throttle_class:
    try:
        throttle = import_class(throttle_class)()
        if not isinstance(throttle, BaseThrottle):
            print "Configured throttle class (%s) is not a BaseThrottle instance, skipping throttling." % throttle_class
            throttle = None
    except ImportError:
        print 'Error importing %s' % throttle_class









__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

from chembl_beaker.beaker import app
from bottle import request, response
from chembl_beaker.beaker.core_apps.standarisation.impl import _break_bonds, _neutralise, _rules, _unsalt, _standardise
import base64

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/break_bonds/<ctab>', method=['OPTIONS', 'GET'], name="break_bonds")
def break_bonds(ctab):
    """
Break covalent bonds between oxygen or nitrogen atoms and Group I and II metal atoms.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/01\_break\_bonds.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/01_break_bonds.html)
CTAB is urlsafe_base64 encoded string containing single molfile or concatenation of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _break_bonds(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/break_bonds', method=['OPTIONS', 'POST'], name="break_bonds")
def break_bonds():
    """
Break covalent bonds between oxygen or nitrogen atoms and Group I and II metal atoms.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/01\_break\_bonds.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/01_break_bonds.html)
CTAB is either single molfile or SDF file.
    """

    data = request.files.values()[0].file.read() if len(request.files) else request.body.read()
    return _break_bonds(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/neutralise/<ctab>', method=['OPTIONS', 'GET'], name="neutralise")
@app.route('/neutralise/<ctab>/<balance>', method=['OPTIONS', 'GET'], name="neutralise")
def neutralise(ctab, balance=False):
    """
Neutralise charges.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/02\_neutralise.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/02_neutralise.html)
CTAB is urlsafe_base64 encoded string containing single molfile or
concatenation of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _neutralise(data, balance)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/neutralise', method=['OPTIONS', 'POST'], name="neutralise")
def neutralise():
    """
Neutralise charges.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/02\_neutralise.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/02_neutralise.html)
CTAB is either single molfile or SDF file.
    """

    balance = int(request.forms.get('balance', False))
    data = request.files.values()[0].file.read() if len(request.files) else request.body.read()
    return _neutralise(data, balance)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/unsalt/<ctab>', method=['OPTIONS', 'GET'], name="unsalt")
def unsalt(ctab):
    """
Remove counterions and solvate components.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/04\_unsalt.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/04_unsalt.html)
CTAB is urlsafe_base64 encoded string containing single molfile or
concatenation of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _unsalt(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/unsalt', method=['OPTIONS', 'POST'], name="unsalt")
def unsalt():
    """
Remove salt/solvates.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/04\_unsalt.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/04_unsalt.html)
CTAB is either single molfile or SDF file.
    """

    data = request.files.values()[0].file.read() if len(request.files) else request.body.read()
    return _unsalt(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/rules/<ctab>', method=['OPTIONS', 'GET'], name="rules")
def rules(ctab):
    """
Apply structure-normalisation transformations.
List of rules and further documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/03\_rules.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/03_rules.html)
CTAB is urlsafe_base64 encoded string containing single molfile or
concatenation of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _rules(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/rules', method=['OPTIONS', 'POST'], name="rules")
def rules():
    """
Apply structure-normalisation transformations.
List of rules and further documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/03\_rules.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/03_rules.html)
CTAB is either single molfile or SDF file.
    """

    data = request.files.values()[0].file.read() if len(request.files) else request.body.read()
    return _rules(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/standardise/<ctab>', method=['OPTIONS', 'GET'], name="standardise")
def standardise(ctab):
    """
Get standardised parent.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/05\_standardise.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/05_standardise.html)
CTAB is urlsafe_base64 encoded string containing single molfile or
concatenation of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _standardise(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/standardise', method=['OPTIONS', 'POST'], name="standardise")
def standardise():
    """
Get standardised parent.
Examples and documentation: [https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/05\_standardise.html](https://wwwdev.ebi.ac.uk/chembl/extra/francis/standardiser/05_standardise.html)
CTAB is either single molfile or SDF file.
    """

    data = request.files.values()[0].file.read() if len(request.files) else request.body.read()
    return _standardise(data)

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

import time
import json
from bottle import request, response
from chembl_beaker.beaker import config
from chembl_beaker.beaker.cache import cache

if not cache:
    print "Caching plugin enabled but no cache backend configured, cashing will be skipped..."

if cache and config.get('clear_cache_on_start', False):
    cache.clear()

#-----------------------------------------------------------------------------------------------------------------------

class Caching(object):
    name = 'caching'
    api = 2

    def apply(self, fn, context):
        def _caching(*args, **kwargs):
            start = time.time()
            if not cache:
                from_cache = False
                res = fn(*args, **kwargs)
            else:
                key = json.dumps(args) + json.dumps(kwargs) + request.method + request.path + request.body.read()
                request.body.seek(0)
                cached_content, content_type = cache.get(key, (None, None))
                if cached_content:
                    if content_type:
                        response.headers['Content-Type'] = content_type
                    from_cache = True
                    res = cached_content
                else:
                    from_cache = False
                    res = fn(*args, **kwargs)
                    if type(res) == str:
                        content_type = response.headers.get('Content-Type')
                        cache.set(key, (res, content_type))
            if config.get('debug', True):
                end = time.time()
                response.headers['X-ChEMBL-in-cache'] = from_cache
                response.headers['X-ChEMBL-retrieval-time'] = end - start
            return res

        return _caching

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

from datetime import datetime
import pytz
import base64
from bottle import request, response
from chembl_beaker.beaker.throttle import throttle
from chembl_beaker.beaker import config
from chembl_beaker.beaker.utils import import_class

DATE_FORMAT = '%y-%m-%d %H%M%S'

AES = None
secret_key = None
try:
    from Crypto.Cipher import AES
    secret_key = config.get('throttling_secret_key')
except ImportError:
    pass

#-----------------------------------------------------------------------------------------------------------------------

def verify_key(key):
    if not (AES or secret_key):
        return False
    aes = AES.new(secret_key, AES.MODE_CBC)

    try:
        now = datetime.utcnow().replace(tzinfo=pytz.utc)
        decoded_key = base64.standard_b64decode(key)
        decrypted_key = aes.decrypt(decoded_key)
        decrypted = filter(bool, decrypted_key.split('\t'))
        dates = [pytz.utc.localize(datetime.strptime(date, DATE_FORMAT)) for date in decrypted]
        if dates[0] < now < dates[1]:
            return True
    except:
        pass

    return False

#-----------------------------------------------------------------------------------------------------------------------

key_verification_function = config.get('throttling_key_verification_function')
if key_verification_function:
    try:
        verified_key = import_class(key_verification_function)
    except ImportError:
        err = 'Error importing function %s' % key_verification_function
        print err
        raise Exception(err)

else:
    if not AES:
        print "API Key verification will be switched off - can't find pycrypto library"
    if not secret_key:
        print "API Key verification will be switched off - no AES key defined"
    verified_key = verify_key

#-----------------------------------------------------------------------------------------------------------------------

def get_identifier(req):
    ip = req.remote_addr
    api_key = req.headers.get('X-ChEMBL-APIKey')
    if api_key:
        if verified_key(api_key):
            return api_key, 'API Key'
        return ip, 'IP (API Key verification failed)'
    return ip, 'IP'

#-----------------------------------------------------------------------------------------------------------------------

class Throttling(object):
    name = 'throttling'
    api = 2

    def apply(self, fn, context):
        def _throttle_check(*args, **kwargs):

            res = fn(*args, **kwargs)
            if not type(res) == str:
                return res
            else:
                identifier, auth_type = get_identifier(request)
                hourly_limit = config.get('throttle_hourly_rate_limit', 60 * 60) if auth_type == 'IP' else \
                                                        config.get('throttle_key_hourly_rate_limit', 60 * 60 * 2)
                daily_limit = config.get('throttle_daily_rate_limit', 60 * 60 * 24) if auth_type == 'IP' else \
                                                    config.get('throttle_key_daily_rate_limit', 60 * 60 * 24 * 2)
                response.headers['X-ChEMBL-Authentication-Type'] = auth_type
                hourly_remaining, daily_remaining = throttle.get_remaining_rates(identifier, auth_type)
                response.headers['X-HourlyRateLimit-Limit'] = hourly_limit
                response.headers['X-DailyRateLimit-Limit'] = daily_limit
                response.headers['X-HourlyRateLimit-Remaining'] = hourly_remaining
                response.headers['X-DailyRateLimit-Remaining'] = daily_remaining
                if not all((hourly_remaining, daily_remaining)):
                    response.status = 429
                    response.body = 'Too many requests, try again later'
                    return response

                throttle.accessed(identifier, auth_type)

                return res

        return _throttle_check

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

from chembl_beaker.beaker import app
from bottle import request
from chembl_beaker.beaker.core_apps.conversions.impl import _ctab2smiles, _smiles2ctab, _inchi2ctab, _ctab2inchi, _inchi2inchiKey, _canonicalize_smiles
import base64

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/ctab2smiles/<ctab>', method=['OPTIONS', 'GET'], name="ctab2smiles")
def ctab2smiles(ctab):
    """
Converts CTAB to SMILES format. CTAB is urlsafe_base64 encoded string containing single molfile or concatenation
of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _ctab2smiles(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/ctab2smiles', method=['OPTIONS', 'POST'], name="ctab2smiles")
def ctab2smiles():
    """
Converts CTAB to SMILES format. CTAB is either single molfile or SDF file.
    """

    data=request.body.read()
    return _ctab2smiles(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/smiles2ctab/<smiles>', method=['OPTIONS', 'GET'], name="smiles2ctab")
def smiles2ctab(smiles):
    """
Converts SMILES to CTAB. This method accepts urlsafe_base64 encoded string containing single or multiple SMILES
optionally containing header line, specific to *.smi format.
    """

    data = base64.urlsafe_b64decode(smiles)
    if not data.startswith('SMILES Name'):
        data = "SMILES Name\n" + data
    return _smiles2ctab(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/smiles2ctab', method=['OPTIONS', 'POST'], name="smiles2ctab")
def smiles2ctab():
    """
Converts SMILES to CTAB. This method accepts single or multiple SMILES or *.smi file.
    """

    data = request.body.read()
    if not data.startswith('SMILES Name'):
        data = "SMILES Name\n" + data
    return _smiles2ctab(data)
#-----------------------------------------------------------------------------------------------------------------------

@app.route('/canonicalizeSmiles/<smiles>', method=['OPTIONS', 'GET'], name="canonicalizeSmiles")
def canonicalizeSmiles(smiles):
    """
Converts SMILES to canonical SMILES. This method accepts urlsafe_base64 encoded string containing single or multiple SMILES
optionally containing header line, specific to *.smi format.
    """

    data = base64.urlsafe_b64decode(smiles)
    if not data.startswith('SMILES Name'):
        data = "SMILES Name\n" + data
    return _canonicalize_smiles(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/canonicalizeSmiles', method=['OPTIONS', 'POST'], name="canonicalizeSmiles")
def canonicalizeSmiles():
    """
Converts SMILES to canonical SMILES. This method accepts single or multiple SMILES or *.smi file.
    """

    data = request.body.read()
    if not data.startswith('SMILES Name'):
        data = "SMILES Name\n" + data
    return _canonicalize_smiles(data)
#-----------------------------------------------------------------------------------------------------------------------

@app.route('/inchi2ctab/<inchi>', method=['OPTIONS', 'GET'], name="inchi2ctab")
def inchi2ctab(inchi):
    """
Converts InChi to CTAB. This method accepts urlsafe_base64 encoded string containing one or multiple InChis.
    """

    inchis = base64.urlsafe_b64decode(inchi)
    return _inchi2ctab(inchis)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/inchi2ctab', method=['OPTIONS', 'POST'], name="inchi2ctab")
def inchi2ctab():
    """
Converts InChi to CTAB. This method accepts one or multiple InChis.
    """

    inchis = request.body.read()
    return _inchi2ctab(inchis)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/ctab2inchi/<ctab>', method=['OPTIONS', 'GET'], name="ctab2inchi")
def ctab2inchi(ctab):
    """
Converts CTAB to InChis. CTAB is urlsafe_base64 encoded string containing single molfile or concatenation
of multiple molfiles.
    """

    data = base64.urlsafe_b64decode(ctab)
    return _ctab2inchi(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/ctab2inchi', method=['OPTIONS', 'POST'], name="ctab2inchi")
def ctab2inchi():
    """
Converts CTAB to InChis. CTAB is either single molfile or SDF file.
    """

    data=request.body.read()
    return _ctab2inchi(data)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/inchi2inchiKey/<inchi>', method=['OPTIONS', 'GET'], name="inchi2inchiKey")
def inchi2inchiKey(inchi):
    """
Converts InChis to InChiKeys. This method accepts urlsafe_base64 encoded string containing one or multiple InChis.
    """

    inchis = base64.urlsafe_b64decode(inchi)
    return _inchi2inchiKey(inchis)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/inchi2inchiKey', method=['OPTIONS', 'POST'], name="inchi2inchiKey")
def inchi2inchiKey():
    """
Converts InChis to InChiKeys. This method accepts one or multiple InChis.
    """

    inchis = request.body.read()
    return _inchi2inchiKey(inchis)

#-----------------------------------------------------------------------------------------------------------------------



# $Id$
#
# Copyright (C) 2006-2011 Greg Landrum
#  All Rights Reserved
#
import os
from chembl_beaker.beaker.draw.molDrawing import MolDrawing,DrawingOptions

def _getCanvas():
  useAGG=False
  useCairo=True
  from chembl_beaker.beaker.draw.cairoCanvas import Canvas
  return useAGG,useCairo,Canvas

def _createCanvas(size):
  useAGG,useCairo,Canvas=_getCanvas()
  from PIL import Image
  img = Image.new("RGBA",size,(0,0,0,0))
  canvas = Canvas(img)
  return img,canvas

def MolToImage(mol, size=(300,300), kekulize=True, wedgeBonds=True,
               fitImage=False, options=None, canvas=None, **kwargs):
  """ returns a PIL image containing a drawing of the molecule

    Keyword arguments:
    kekulize -- run kekulization routine on input `mol` (default True)
    size -- final image size, in pixel (default (300,300))
    wedgeBonds -- draw wedge (stereo) bonds (default True)
    highlightAtoms -- list of atoms to highlight (default [])
    highlightMap -- dictionary of (atom, color) pairs (default None)
    highlightBonds -- list of bonds to highlight (default [])
  """
  if not mol:
    raise ValueError('Null molecule provided')
  if canvas is None:
    img,canvas=_createCanvas(size)
  else:
    img=None
    
  if options is None:
    options = DrawingOptions()
  if fitImage:
      options.dotsPerAngstrom = int(min(size) / 10)
  options.wedgeDashedBonds = wedgeBonds
  drawer = MolDrawing(canvas=canvas,drawingOptions=options)

  if kekulize:
    from rdkit import Chem
    mol = Chem.Mol(mol.ToBinary())
    Chem.Kekulize(mol)
    
  if not mol.GetNumConformers():
    from rdkit.Chem import AllChem
    AllChem.Compute2DCoords(mol)
  
  if 'legend' in kwargs:
    legend = kwargs['legend']
    del kwargs['legend']
  else:
    legend=''

  drawer.AddMol(mol,**kwargs)

  if legend:
    from chembl_beaker.beaker.Draw.MolDrawing import Font
    bbox = drawer.boundingBoxes[mol]
    pos = size[0]/2,int(.94*size[1]),0 # the 0.94 is extremely empirical
    # canvas.addCanvasPolygon(((bbox[0],bbox[1]),(bbox[2],bbox[1]),(bbox[2],bbox[3]),(bbox[0],bbox[3])),
    #                         color=(1,0,0),fill=False,stroke=True)
    # canvas.addCanvasPolygon(((0,0),(0,size[1]),(size[0],size[1]),(size[0],0)   ),
    #                         color=(0,0,1),fill=False,stroke=True)
    font=Font(face='sans',size=12)
    canvas.addCanvasText(legend,pos,font)

  if kwargs.get('returnCanvas',False):
    return img,canvas,drawer
  else:
    canvas.flush()
    return img

def MolToFile(mol,fileName,size=(300,300),kekulize=True, wedgeBonds=True,
              imageType=None, fitImage=False, options=None, **kwargs):
  """ Generates a drawing of a molecule and writes it to a file
  """
  # original contribution from Uwe Hoffmann
  if not fileName:
    raise ValueError('no fileName provided')
  if not mol:
    raise ValueError('Null molecule provided')

  if imageType is None:
    imageType=os.path.splitext(fileName)[1][1:]

  if options is None:
    options = DrawingOptions()
  useAGG,useCairo,Canvas = _getCanvas()
  if fitImage:
      options.dotsPerAngstrom = int(min(size) / 10)
  options.wedgeDashedBonds = wedgeBonds
  if useCairo or useAGG:
    canvas = Canvas(size=size,imageType=imageType,
                              fileName=fileName)
  else:
    options.radicalSymbol = '.' #<- the sping canvas doesn't support unicode well
    canvas = Canvas(size=size,name=fileName,imageType=imageType)
  drawer = MolDrawing(canvas=canvas,drawingOptions=options)
  if kekulize:
    from rdkit import Chem
    mol = Chem.Mol(mol.ToBinary())
    Chem.Kekulize(mol)
    
  if not mol.GetNumConformers():
    from rdkit.Chem import AllChem
    AllChem.Compute2DCoords(mol)

  drawer.AddMol(mol,**kwargs)
  if useCairo or useAGG:
    canvas.flush()
  else:
    canvas.save()

def MolToImageFile(mol,filename,size=(300,300),kekulize=True, wedgeBonds=True,
                   **kwargs):
  """  DEPRECATED:  please use MolToFile instead

  """
  img = MolToImage(mol,size=size,kekulize=kekulize,wedgeBonds=wedgeBonds,**kwargs)
  img.save(filename)
    
tkRoot=None
tkLabel=None
tkPI=None
def ShowMol(mol,size=(300,300),kekulize=True,wedgeBonds=True,
            title='RDKit Molecule',**kwargs):
  """ Generates a picture of a molecule and displays it in a Tkinter window
  """
  global tkRoot,tkLabel,tkPI
  import Tkinter
  try:
    import ImageTk
  except ImportError:
    from PIL import ImageTk

  img = MolToImage(mol,size,kekulize,wedgeBonds,**kwargs)

  if not tkRoot:
    tkRoot = Tkinter.Tk()
    tkRoot.title(title)
    tkPI = ImageTk.PhotoImage(img)
    tkLabel = Tkinter.Label(tkRoot,image=tkPI)
    tkLabel.place(x=0,y=0,width=img.size[0],height=img.size[1])
  else:
    tkPI.paste(img)
  tkRoot.geometry('%dx%d'%(img.size))

def calcAtomGaussians(mol,a=0.03,step=0.02,weights=None):
  import numpy
  from matplotlib import mlab
  x = numpy.arange(0,1,step)
  y = numpy.arange(0,1,step)
  X,Y = numpy.meshgrid(x,y)
  if weights is None:
    weights=[1.]*mol.GetNumAtoms()
  Z = mlab.bivariate_normal(X,Y,a,a,mol._atomPs[0][0], mol._atomPs[0][1])*weights[0]
  for i in range(1,mol.GetNumAtoms()):
    Zp = mlab.bivariate_normal(X,Y,a,a,mol._atomPs[i][0], mol._atomPs[i][1])
    Z += Zp*weights[i]
  return X,Y,Z
  

def MolsToImage(mols, subImgSize=(200,200),legends=None,**kwargs):
  """ 
  """
  try:
    import Image
  except ImportError:
    from PIL import Image
  if legends is None: legends = [None]*len(mols)
  res = Image.new("RGBA",(subImgSize[0]*len(mols),subImgSize[1]))
  for i,mol in enumerate(mols):
    res.paste(MolToImage(mol,subImgSize,legend=legends[i],**kwargs),(i*subImgSize[0],0))
  return res


def MolsToGridImage(mols,molsPerRow=3,subImgSize=(200,200),legends=None,
                    highlightAtomLists=None,**kwargs):
  """ 
  """
  try:
    import Image
  except ImportError:
    from PIL import Image
  if legends is None: legends = [None]*len(mols)

  nRows = len(mols)//molsPerRow
  if len(mols)%molsPerRow : nRows+=1
    
  res = Image.new("RGBA",(molsPerRow*subImgSize[0],nRows*subImgSize[1]),(255,255,255,0))
  for i,mol in enumerate(mols):
    row = i//molsPerRow
    col = i%molsPerRow
    highlights=None
    if highlightAtomLists and highlightAtomLists[i]:
      highlights=highlightAtomLists[i]
    res.paste(MolToImage(mol,subImgSize,legend=legends[i],highlightAtoms=highlights,
                         **kwargs),(col*subImgSize[0],row*subImgSize[1]))
  return res

def ReactionToImage(rxn, subImgSize=(200,200),**kwargs):
  """ 
  """
  try:
    import Image
  except ImportError:
    from PIL import Image

  mols = []
  for i in range(rxn.GetNumReactantTemplates()):
    tmpl=rxn.GetReactantTemplate(i)
    tmpl.UpdatePropertyCache(False)
    mols.append(tmpl)
  mols.append(None)
  for i in range(rxn.GetNumProductTemplates()):
    tmpl = rxn.GetProductTemplate(i)
    tmpl.UpdatePropertyCache(False)
    mols.append(tmpl)

  res = Image.new("RGBA",(subImgSize[0]*len(mols),subImgSize[1]),(255,255,255,0))
  for i,mol in enumerate(mols):
    if mol is not None:
      nimg = MolToImage(mol,subImgSize,kekulize=False,**kwargs)
    else:
      nimg,canvas = _createCanvas(subImgSize)
      p0 = (10,subImgSize[1]//2)
      p1 = (subImgSize[0]-10,subImgSize[1]//2)
      p3 = (subImgSize[0]-20,subImgSize[1]//2-10)
      p4 = (subImgSize[0]-20,subImgSize[1]//2+10)
      canvas.addCanvasLine(p0,p1,lineWidth=2,color=(0,0,0))
      canvas.addCanvasLine(p3,p1,lineWidth=2,color=(0,0,0))
      canvas.addCanvasLine(p4,p1,lineWidth=2,color=(0,0,0))
      if hasattr(canvas,'flush'):
        canvas.flush()
      else:
        canvas.save()        
    res.paste(nimg,(i*subImgSize[0],0))
  return res




__author__ = 'mnowotka'

#-----------------------------------------------------------------------------------------------------------------------

try:
    import cPickle as pickle
except ImportError:
    import pickle
from threading import Thread
from Queue import Queue
import base64
import sqlite3
from datetime import datetime
from chembl_beaker.beaker.cache.backends.base import BaseCache
from chembl_beaker.beaker import config
import pytz

#-----------------------------------------------------------------------------------------------------------------------

def value_to_db_datetime(value):
    if value is None:
        return None
    if value.tzinfo is not None and value.tzinfo.utcoffset(value) is not None:
        value = value.astimezone(pytz.utc).replace(tzinfo=None)
    return value

#-----------------------------------------------------------------------------------------------------------------------

class MultiThreadOK(Thread):
    def __init__(self, db):
        super(MultiThreadOK, self).__init__()
        self.db=db
        self.reqs=Queue()
        self.start()
    def run(self):
        cnx = sqlite3.connect(self.db)
        cursor = cnx.cursor()
        while True:
            req, arg, res = self.reqs.get()
            if req=='--close--': break
            cursor.execute(req, arg)
            if res:
                for rec in cursor:
                    res.put(rec)
                res.put('--no more--')
        cnx.close()
    def execute(self, req, arg=None, res=None):
        self.reqs.put((req, arg or tuple(), res))
    def select(self, req, arg=None):
        res=Queue()
        self.execute(req, arg, res)
        while True:
            rec=res.get()
            if rec=='--no more--': break
            yield rec
    def close(self):
        self.execute('--close--')

#-----------------------------------------------------------------------------------------------------------------------

class SQLiteCache(BaseCache):
    def __init__(self):
        super(SQLiteCache, self).__init__()
        self._table = config.get('sqlite_cache_table', 'beaker_cache')
        self.sql = MultiThreadOK(config.get('sqlite_dbfile', ':memory:'))
        self.sql.execute("CREATE TABLE IF NOT EXISTS %s (cache_key varchar(255) PRIMARY KEY NOT NULL, value text, expires datetime NOT NULL)" % self._table)

#-----------------------------------------------------------------------------------------------------------------------

    def get(self, key, default=None):
        key = self.make_key(key)

        try:
            row = next(self.sql.select("SELECT cache_key, value, expires FROM %s "
                               "WHERE cache_key = ?" % self._table, [key]))
        except StopIteration:
            row = None

        if row is None:
            return default
        now = datetime.utcnow().replace(tzinfo=pytz.utc)
        expires = pytz.utc.localize(datetime.strptime(row[2], '%Y-%m-%d %H:%M:%S'))

        if expires < now:
            self.sql.execute("DELETE FROM %s "
                               "WHERE cache_key = ?" % self._table, [key])
            return default
        value = row[1]
        return pickle.loads(base64.b64decode(value))

#-----------------------------------------------------------------------------------------------------------------------

    def set(self, key, value):
        key = self.make_key(key)
        self._base_set('set', key, value, self.get_backend_timeout())

#-----------------------------------------------------------------------------------------------------------------------

    def add(self, key, value):
        key = self.make_key(key)
        return self._base_set('add', key, value, self.get_backend_timeout())

#-----------------------------------------------------------------------------------------------------------------------

    def _base_set(self, mode, key, value, timeout):

        row = next(self.sql.select("SELECT COUNT(*) FROM %s" % self._table))
        num = row[0]
        now = datetime.utcnow().replace(tzinfo=pytz.utc)
        now = now.replace(microsecond=0)
        if timeout is None:
            exp = datetime.max
        else:
            exp = datetime.fromtimestamp(timeout)
        exp = exp.replace(microsecond=0)
        if num > self._max_entries:
            self._cull(now)
        pickled = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
        b64encoded = base64.b64encode(pickled)

        # Note: typecasting for datetimes is needed by some 3rd party
        # database backends. All core backends work without typecasting,
        # so be careful about changes here - test suite will NOT pick
        # regressions.

        try:
            result = next(self.sql.select("SELECT cache_key, expires FROM %s WHERE cache_key = ?" % self._table, [key]))
        except StopIteration:
            result = None
        if result:
            current_expires = result[1]

        exp = value_to_db_datetime(exp)
        if result and (mode == 'set' or (mode == 'add' and current_expires < now)):
            self.sql.execute("UPDATE %s SET value = ?, expires = ? WHERE cache_key = ?" % self._table,
                           (b64encoded, exp, key))
        else:
            self.sql.execute("INSERT INTO %s (cache_key, value, expires) VALUES (?, ?, ?)" % self._table,
                           (key, b64encoded, exp))

            return True

#-----------------------------------------------------------------------------------------------------------------------

    def delete(self, key):
        key = self.make_key(key)
        self.sql.execute("DELETE FROM %s WHERE cache_key = ?" % self._table, [key])

#-----------------------------------------------------------------------------------------------------------------------

    def _cull(self, now):
        if self._cull_frequency == 0:
            self.clear()
        else:
            # When USE_TZ is True, 'now' will be an aware datetime in UTC.
            now = now.replace(tzinfo=None)
            self.sql.execute("DELETE FROM %s WHERE expires < ?" % self._table,
                           [value_to_db_datetime(now)])
            row = next(self.sql.select("SELECT COUNT(*) FROM %s" % self._table))
            num = row[0]
            if num > self._max_entries:
                cull_num = num // self._cull_frequency
                row = next(self.sql.select(
                    "SELECT cache_key FROM %s ORDER BY cache_key LIMIT 1 OFFSET %s" % (self._table,
                    cull_num)))
                self.sql.execute("DELETE FROM %s "
                               "WHERE cache_key < ?" % self._table,
                               [row[0]])

#-----------------------------------------------------------------------------------------------------------------------

    def clear(self):
        self.sql.execute('DELETE FROM %s' % self._table)

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

from bottle import request, response
import json

from chembl_beaker.beaker import app
from chembl_beaker.beaker.core_apps.marvin.impl import _clean, _stereoInfo, _molExport, _hydrogenize

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/clean', method=['OPTIONS', 'POST'], name="clean")
def clean():
    """
Implements Marvin 4 js clean2D/3D web service. Recomputes 2D/3D coordinates of given compound. The compound is in
Marvin's *.mrv format. Dim is an optional parameter specifying if 2D or 3D coordinates should be computed
    """

    params = json.loads(request.body.read())
    structure = params['structure']
    dim = params.get('dim', 2)
    response.content_type = 'text/plain'
    return _clean(structure, dim)

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/hydrogenize', method=['OPTIONS', 'POST'], name="hydrogenize")
def hydrogenize():
    """
Implements Marvin 4 js Hydrogenize web service. Adds or removes hydrogen atoms. The compound is in
Marvin's *.mrv format. Dim is an optional parameter specifying if 2D or 3D coordinates should be computed
    """

    params = json.loads(request.body.read())
    structure = params['structure']
    input_format = params.get('inputFormat', 'mrv')
    parameters = params.get('parameters')
    method = 'hydrogenize'
    if parameters:
        method = parameters.get('method', 'hydrogenize')
        method = method.lower()
    response.content_type = 'text/plain'
    return _hydrogenize(_molExport(structure, input=input_format, output='mol')["structure"], method == 'hydrogenize')

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/cipStereoInfo', method=['OPTIONS', 'POST'], name="cipStereoInfo")
def stereoInfo():
    """
Implements Marvin 4 js Stereo Info web service. Marks possible stereocenters, R/S chirality for atoms and cistrans
bonds.
    """

    params = json.loads(request.body.read())
    structure = params['structure']
    response.content_type = 'application/json'
    return json.dumps(_stereoInfo(structure))

#-----------------------------------------------------------------------------------------------------------------------

@app.route('/molExport' , method=['OPTIONS', 'POST'], name="molExport")
def molExport():
    """
Implements Marvin 4 js MolConvert web service. Converts compound from one format to another, including Marvin's
*.mrv.
    """

    params = json.loads(request.body.read())
    structure = params['structure']
    input_f = params.get('inputFormat')
    output_f = params.get('parameters','mrv')

    res = _molExport(structure, input=input_f, output=output_f)
    response.content_type = 'application/json'
    return json.dumps(res)

#-----------------------------------------------------------------------------------------------------------------------
